"""
The PATO GUI for launching Maven builds based on PATO.
"""

# Python modules
import sys
import os
import argparse
import subprocess
import re
from pathlib import Path
import logging
from shutil import which
# from pkg_resources import packaging
import pkg_resources


# items to test
__all__ = ['db_order', 'initialize', 'check_environment', 'process_POM']


logger = None


def db_order(db):
    for i, e in enumerate(['dev', 'tst', 'test', 'acc', 'prod', 'prd']):
        if e in db.lower():
            return i
    return db.lower()


def initialize():
    global logger

    argv = [argc for argc in sys.argv[1:] if argc != '--']

    parser = argparse.ArgumentParser(description='Setup logging')
    parser.add_argument('-d', dest='debug', action='store_true', help='Enable debugging')
    parser.add_argument('--db-config-dir', help='The database configuration directory')
    parser.add_argument('file', nargs='?', help='The POM file')
    args, rest = parser.parse_known_args(argv)
    if args.db_config_dir:
        args.db_config_dir = os.path.abspath(args.db_config_dir)
    if args.file:
        args.file = os.path.abspath(args.file)
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG if args.debug else logging.INFO)
    logger = logging.getLogger()
    if len(rest) == 0 and args.file:
        args.mvnd = 'mvnd' in check_environment()
    else:
        args.mvnd = False
    if '-d' in argv:
        argv.remove('-d')
    logger.debug('argv: %s; logger: %s; args: %s' % (argv, logger, args))
    return argv, logger, args


def check_environment():
    programs = [
        ['mvn', '-version', '3.3.1', r'Apache Maven ([0-9.]+)', True, True],
        ['perl', '--version', '5.16.0', r'\(v([0-9.]+)\)', True, True],
        ['sql', '-V', '18.0.0.0', r'SQLcl: Release ([0-9.]+)', True, True],
        ['java', '-version', '1.8.0', r'(?:java|openjdk) version "([0-9.]+).*"', False, True],  # version is printed to stderr (!#$?)
        ['javac', '-version', '1.8.0', r'javac ([0-9.]+)', True, True],
        ['mvnd', '--version', '0.8.0', r'mvnd ([0-9.]+)', True, False],  # Maven daemon may be there or not
    ]
    programs_found = []

    for i, p in enumerate(programs):
        # p[0]: program
        # p[1]: command line option to get the version
        # p[2]: minimum version
        # p[3]: regular expression to parse for version
        # p[4]: print stdout (True) or stderr (False)?
        # p[5]: program mandatory?
        proc = subprocess.run(p[0] + ' ' + p[1], shell=True, capture_output=True, text=True)
        assert not (p[5]) or proc.returncode == 0, proc.stderr

        if proc.returncode == 0:
            logger.debug('proc: {}'.format(proc))
            expected_version = p[2]
            regex = p[3]
            output = proc.stdout if p[4] else proc.stderr
            m = re.search(regex, output)
            assert m, 'Could not find {} in {}'.format(regex, output)
            actual_version = m.group(1)
            assert pkg_resources.packaging.version.parse(actual_version) >= pkg_resources.packaging.version.parse(expected_version), f'Version of program "{p[0]}" is "{actual_version}" which is less than the expected version "{expected_version}"'
            logger.info('Version of "{}" is "{}" and its location is "{}"'.format(p[0], actual_version, os.path.dirname(which(p[0]))))
            programs_found.append(p[0])
        else:
            logger.info('Command "{0}" failed: {1}'.format(p[0] + ' ' + p[1], proc.stderr))

    return programs_found


def process_POM(pom_file, db_config_dir):
    """
    Process a single POM file and setup the GUI.
    The POM file must be either based on an PATO parent POM for the database or Apex.
    """
    def determine_POM_settings(pom_file, db_config_dir):
        properties = {}
        profiles = set()

        cmd = f"mvn --file {pom_file} -N help:all-profiles -Pconf-inquiry compile"
        if db_config_dir:
            cmd += f" -Ddb.config.dir={db_config_dir}"
        mvn = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, text=True)
        stdout, stderr = mvn.communicate()

        if mvn.returncode == 0:
            pass
        else:
            returncode = mvn.returncode
            error = ''
            for ch in stderr:
                error += ch
            raise Exception(f'The command "{cmd}" failed with return code {returncode} and error:\n{error}')

        # Profile Id: db-install (Active: false , Source: pom)
        line = ''
        for ch in stdout:
            if ch != "\n":
                line += ch
            else:
                logger.debug("line: %s" % (line))
                m = re.search(r"Profile Id: ([a-zA-Z0-9_.-]+) \(Active: .*, Source: pom\)", line)
                if m:
                    logger.debug("adding profile: %s" % (m.group(1)))
                    profiles.add(m.group(1))
                else:
                    # GJP 2023-09-06 https://github.com/paulissoft/pato-gui/issues/8
                    # Change re.match() into re.search() so we can match not only from the beginning but also in the middle.
                    m = re.search(r'\[echoproperties\] ([a-zA-Z0-9_.-]+)=(.+)$', line)
                    if m:
                        logger.debug("adding property %s = %s" % (m.group(1), m.group(2)))
                        properties[m.group(1)] = m.group(2)
                line = ''
        return properties, profiles

    logger.debug('process_POM()')
    properties, profiles = determine_POM_settings(pom_file, db_config_dir)
    apex_profiles = ['apex-seed-publish', 'apex-export', 'apex-import']
    db_profiles = ['db-info', 'db-install', 'db-code-check', 'db-test', 'db-generate-ddl-full', 'db-generate-ddl-incr']
    if profiles.issuperset(set(apex_profiles)):
        profiles = apex_profiles
    elif profiles.issuperset(set(db_profiles)):
        profiles = db_profiles
    else:
        raise Exception('Profiles (%s) must be a super set of either the Apex (%s) or database (%s) profiles' % (profiles, set(apex_profiles), set(db_profiles)))
    if not db_config_dir:
        # C\:\\dev\\bc\\oracle-tools\\conf\\src => C:\dev\bc\oracle-tools\conf\src =>
        db_config_dir = properties.get('db.config.dir', '').replace('\\:', ':').replace('\\\\', '\\')
    assert db_config_dir, 'The property db.config.dir must have been set in order to choose a database (on of its subdirectories)'
    logger.debug('db_config_dir: ' + db_config_dir)

    p = Path(db_config_dir)
    dbs = []
    try:
        dbs = [d.name for d in filter(Path.is_dir, p.iterdir())]
    except Exception:
        pass
    assert len(dbs) > 0, 'The directory %s must have subdirectories, where each one contains information for one database (and Apex) instance' % (properties['db.config.dir'])

    db_proxy_username = properties.get('db.proxy.username', '')
    db_username = properties.get('db.username', '')
    assert db_proxy_username or db_username, f'The database acount (Maven property db.proxy.username {db_proxy_username} or db.username {db_username}) must be set'

    logger.debug('return: (%s, %s, %s, %s, %s)' % (db_config_dir, dbs, profiles, db_proxy_username, db_username))
    return db_config_dir, dbs, profiles, db_proxy_username, db_username



from utils import about
from utils.pom import db_order


def test_about():
    expected = {'__title__', '__author__', '__email__', '__version_info__', '__version__', '__license__', '__copyright__', '__url__', '__help_url__'}
    actual = set(dir(about))
    assert actual.issuperset(expected), f'The actual names of the about module are {actual} but expected are {expected}'


def test_db_order():
    dbs = ['prd', 'tst', 'acc', 'dev']
    dbs_sorted_actual = sorted(dbs, key=db_order)
    dbs_sorted_expected = ['dev', 'tst', 'acc', 'prd']
    assert dbs_sorted_actual == dbs_sorted_expected
    dbs = ['d', 'a', 'b', 'c']
    dbs_sorted_actual = sorted(dbs, key=db_order)
    dbs_sorted_expected = ['a', 'b', 'c', 'd']
    assert dbs_sorted_actual == dbs_sorted_expected


if __name__ == '__main__':
    test_about()
    test_db_order()



# -*- coding: utf-8 -*-
__title__ = 'Paulissoft Application Tools for Oracle (PATO) GUI'
__package_name__ = 'pato-gui'
__author__ = "Gert-Jan Paulissen"
__description__ = ' '.join('''
Paulissoft Application Tools for Oracle (PATO) GUI.
'''.strip().split())
__email__ = "paulissoft@gmail.com"
__version_info__ = ('3', '0', '0')
__version__ = '.'.join(__version_info__)
__license__ = "MIT License"
__copyright__ = 'Copyright (c) 2021-2023 Gert-Jan Paulissen'
__url__ = "https://github.com/paulissoft/pato-gui"
__help_url__ = "https://paulissoft.github.io/pato-gui"

if __name__ == '__main__':
    print(__version__)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Setup
"""

import sys
from setuptools import find_packages
from setuptools.command.test import test as TestCommand
from distutils.core import setup

# To prevent importing about and thereby breaking the coverage info we use this
# exec hack
about = {}
with open('src/utils/about.py', mode='r', encoding="utf-8") as fp:
    exec(fp.read(), about)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to pytest')]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


install_requires = [
    'bs4',
]

tests_require = [
    'pytest',
    'pytest-cov',
    'pytest-flakes',
    'pytest-pycodestyle',
    'flake8',
]


if sys.argv[-1] == 'info':
    for k, v in about.items():
        print('%s: %s' % (k, v))
    sys.exit()


if __name__ == '__main__':
    with open('README.md', mode='r', encoding="utf-8") as f:
        readme = f.read()

    with open('CHANGELOG.md', mode='r', encoding="utf-8") as f:
        changes = f.read()

    setup(
        name=about['__package_name__'],
        version=about['__version__'],
        author=about['__author__'],
        author_email=about['__email__'],
        description=about['__description__'],
        url=about['__url__'],
        license=about['__license__'],
        keywords=["Oracle", "PATO", "GUI"],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        #namespace_packages=[about['__package_name__']],
        long_description=readme + changes,
        long_description_content_type='text/markdown',
        #include_package_data=True,
        install_requires=install_requires,
        tests_require=tests_require,
        setup_requires=[
            'setuptools>=42',
        ],
        zip_safe=True,
        cmdclass={'test': PyTest},
        extras_require={'test': tests_require},
        entry_points = {
            'console_scripts': [
                'pato-gui = pato_gui:main',
            ],
        },
        classifiers=[
            'Development Status :: 6 - Mature',
            'Programming Language :: Python :: 3',
            'Natural Language :: English',
            'Topic :: Utilities',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'License :: OSI Approved :: ' + about['__license__']
        ],
    )
